import gc
import numpy as np
from .surrogate import SurrogateOptimiser


class BarycentreSurrogateOptimiser(SurrogateOptimiser):
    def __init__(self, func, surr, acqf, **kwargs):
        super(BarycentreSurrogateOptimiser, self).__init__(func, surr, acqf, **kwargs)
        
        self.surr.verbose   = self.verbose
        self.acqf.verbose   = self.verbose
        self.candidates     = None
        self.surr_n_epochs  = surr.n_epochs
    

    def predict(self, X, y, n_pred=1):
        
        if self.surr.n_epochs == 1:
            del self.acqf.surr
            gc.collect()

        preds = []

        N = int(self.surr_n_epochs * self.acqf.wp)

        self.surr.n_epochs = N
        self.surr.fit(X, y, epoch=0)
        
        for i in range(self.surr_n_epochs - N):
            self.surr.n_epochs = 1
            self.surr.fit(X, y, epoch=1)
            preds.append(self.surr)
        
        self.acqf.surr = preds
        
        self.acqf.y = y.reshape(y.shape[0]//self.func.n_obs, self.func.n_obs)
        self.acqf.conds = X[::self.func.n_obs]
        
        next_x = []
        for _ in range(n_pred):
            next_x.append(self.numerical_search([1]))
        
        del preds
        gc.collect()

        return np.array(next_x)


class EnsembleBarycentreSurrogateOptimiser(SurrogateOptimiser):
    def __init__(self, func, surr, acqf, **kwargs):
        super(EnsembleBarycentreSurrogateOptimiser, self).__init__(func, surr, acqf, **kwargs)
        
        for surr in self.surr:
            surr.verbose   = self.verbose
        self.acqf.verbose   = self.verbose
        self.candidates     = None
    

    def predict(self, X, y, n_pred=1):
        
        surrs = []
        
        for surr in self.surr:
            
            surr.fit(X, y)
            surrs.append(surr)
        
        self.acqf.surr = surrs
        
        self.acqf.y = y.reshape(y.shape[0]//self.func.n_obs, self.func.n_obs)
        self.acqf.conds = X[::self.func.n_obs]
        
        next_x = []
        for _ in range(n_pred):
            next_x.append(self.numerical_search([1]))
        
        del surrs
        gc.collect()

        return np.array(next_x)


from .base import Surrogate

import gc
import torch
import gpytorch
from tqdm import tqdm
from gpytorch.models.deep_gps import DeepGP, DeepGPLayer
from gpytorch.models import AbstractVariationalGP
from gpytorch.means import ConstantMean, LinearMean
from gpytorch.kernels import RBFKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.distributions import MultivariateNormal
from gpytorch.mlls import VariationalELBO, DeepApproximateMLL
from gpytorch.variational import CholeskyVariationalDistribution, VariationalStrategy

device = 'cuda' if torch.cuda.is_available() else 'cpu'
torch.set_default_device(device)

class DGP(Surrogate):
    def __init__(self, **kwargs):
        super(DGP, self).__init__()

        self.name         = 'DGP'
        self.model        = kwargs['model'] if 'model' in kwargs else None
        self.n_epochs     = kwargs['n_epochs'] if 'n_epochs' in kwargs else 200
        self.lr           = kwargs['lr'] if 'lr' in kwargs else 1e-1
        self.verbose      = kwargs['verbose'] if 'verbose' in kwargs else 1
        self.num_inducing = kwargs['num_inducing'] if 'num_inducing' in kwargs else 64
        self.hidden_size  = kwargs['hidden_size'] if 'hidden_size' in kwargs else 128
        self.actf         = kwargs['actf'] if 'actf' in kwargs else torch.tanh
        self.means        = kwargs['means'] if 'means' in kwargs else ['linear', 'linear']
        self.scale        = kwargs['scale'] if 'scale' in kwargs else True

        self.gen = torch.Generator() # for reproducibility
        self.gen.manual_seed(2208060503)
    
    def fit(self, X, y, epoch=None):
        
        if not epoch:
            del self.model
            gc.collect()

            self.model = DeepGPModel(
                in_dim       = X.shape[1],
                hidden_size  = self.hidden_size,
                num_inducing = self.num_inducing,
                actf         = self.actf,
                means        = self.means,
                gen          = self.gen
            )
        
        X = torch.tensor(X).float()
        y = torch.tensor(y).float().squeeze()
        
        if self.scale:
            self.y_mu, self.y_std = y.mean(), y.std()
            y = (y - self.y_mu) / (self.y_std + 1e-8)
        
        self.model.train()
        self.model.likelihood.train()
        
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)

        mll = DeepApproximateMLL(VariationalELBO(self.model.likelihood, self.model, num_data=y.shape[0]))

        if self.verbose > 1:
            pbar = tqdm(range(self.n_epochs), leave=False)
        else:
            pbar = range(self.n_epochs)
        
        for epoch in pbar:
            optimizer.zero_grad()
            output = self.model(X)
            loss = -mll(output, y)
            loss.mean().backward()
            optimizer.step()
            
            if self.verbose > 1:
                pbar.set_description(f'Epoch {epoch + 1}/{self.n_epochs} - Loss: {loss.mean().item():.3f}')

    
    def predict(self, X):

        self.model.eval()
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            observed_pred = self.model.likelihood(self.model(torch.tensor(X).float()))
            mean = observed_pred.mean[0, 0, :]
            std = torch.sqrt(observed_pred.variance)[0, 0, :]
        
        if self.scale:
            # print(mean.shape, std.shape, self.y_mu.shape, self.y_std.shape)
            mean += self.y_mu
            std *= self.y_std
        
        return mean.detach().numpy(), std.detach().numpy()


class SingleLayerGP(AbstractVariationalGP):
    def __init__(self, inducing_points, mean='const'):
        variational_distribution = CholeskyVariationalDistribution(inducing_points.size(0))
        variational_strategy = VariationalStrategy(self, inducing_points, variational_distribution)
        super().__init__(variational_strategy)
        self.mean_module = ConstantMean() if mean == 'const' else  LinearMean(inducing_points.size(1))
        self.covar_module = ScaleKernel(RBFKernel(ard_num_dims=inducing_points.size(1))) 

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class DeepLayer(DeepGPLayer):
    def __init__(self, input_dims, output_dims, inducing_points, mean_type='constant'):
        if output_dims is None:
            batch_shape = torch.Size([])
        else:
            batch_shape = torch.Size([output_dims])

        variational_distribution = CholeskyVariationalDistribution(
            inducing_points.size(0),
            batch_shape=batch_shape
        )

        variational_strategy = VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True
        )

        super(DeepLayer, self).__init__(variational_strategy, input_dims, output_dims)

        if mean_type == 'constant':
            self.mean_module = ConstantMean(batch_shape=batch_shape)
        else:
            self.mean_module = LinearMean(input_dims)

        self.covar_module = ScaleKernel(
            RBFKernel(batch_shape=batch_shape, ard_num_dims=input_dims),
            batch_shape=batch_shape, ard_num_dims=input_dims
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class DeepGPModel(DeepGP):
    def __init__(self, in_dim, out_dim=None, hidden_size=16, num_inducing=22, layers=['deep', 'deep'], means=['constant', 'constant'], actf=torch.tanh, gen=None):
        super().__init__()
        
        self.actf = actf
        
        inducing_points = torch.rand(num_inducing, in_dim, generator=gen)
        output_inducing = torch.rand(num_inducing, hidden_size if layers[1]=='deep' else 1, generator=gen)
        
        self.input_layer = DeepLayer(in_dim, hidden_size, inducing_points, mean_type=means[0]) if layers[0]=='deep' else SingleLayerGP(inducing_points)
        self.output_layer =  DeepLayer(hidden_size, None, output_inducing, mean_type=means[1]) if layers[1]=='deep' else SingleLayerGP(output_inducing)
        
        self.likelihood = GaussianLikelihood()

    def forward(self, x):
        hidden_rep = self.input_layer(x).mean
        if self.actf is not None:
            hidden_rep = self.actf(hidden_rep)
        output = self.output_layer(hidden_rep)
        return output



import numpy as np
import pytest

from waggon.functions import Function
from waggon.functions import three_hump_camel
from waggon.functions import ackley
from waggon.functions import rosenbrock
from waggon.functions import submanifold_rosenbrock

from waggon.functions import FunctionV2
from waggon.functions import ThreeHumpCamel
from waggon.functions import Ackley
from waggon.functions import Rosenbrock
from waggon.functions import SubmanifoldRosenbrock
from waggon.functions import StyblinskyTang
from waggon.functions import Levi
from waggon.functions import Himmelblau
from waggon.functions import Holder

from utils import check_func_call_dims
from utils import check_func_sample_dims
from utils import check_func_log_transform


def _test_func(func, func_log):
    is_old_api = True
    if isinstance(func, FunctionV2):
        is_old_api = False
    
    check_func_call_dims(func, func_log, is_old_api)
    check_func_sample_dims(func, func_log, is_old_api)
    check_func_log_transform(func, func_log, is_old_api)


@pytest.mark.parametrize(
    "func, func_log", [
        # Old Three Hump Camel
        (three_hump_camel(log_transform=False), three_hump_camel(log_transform=True)),
        # New Three Hump Camel
        (ThreeHumpCamel(), ThreeHumpCamel(log_transform=True)),
    ]
)
def test_thc(func, func_log):
    _test_func(func, func_log)


@pytest.mark.parametrize(
    "func, func_log", [
        # Old Ackley
        (ackley(dim=2, log_transform=False), ackley(dim=2)),
        (ackley(dim=3, log_transform=False), ackley(dim=3)),
        (ackley(dim=5, log_transform=False), ackley(dim=5)),
        (ackley(dim=10, log_transform=False), ackley(dim=10)),
        # New Ackley
        (Ackley(dim=2), Ackley(dim=2, log_transform=True)),
        (Ackley(dim=3), Ackley(dim=3, log_transform=True)),
        (Ackley(dim=5), Ackley(dim=5, log_transform=True)),
        (Ackley(dim=10), Ackley(dim=10, log_transform=True)),
    ]
)
def test_ackley(func, func_log):
    _test_func(func, func_log)


@pytest.mark.parametrize(
    "func, func_log", [
        # Old API Rosenbrock
        (
            rosenbrock(dim=5, log_transform=False),
            rosenbrock(dim=5)
        ),
        (
            rosenbrock(dim=20, log_transform=False),
            rosenbrock(dim=20)
        ),
        # New API Rosenbrock
        (
            Rosenbrock(dim=7),
            Rosenbrock(dim=7, log_transform=True)
        ),
        (
            Rosenbrock(dim=20),
            Rosenbrock(dim=20, log_transform=True)
        )
    ]
)
def test_rosenbrock(func, func_log):
    _test_func(func, func_log)



@pytest.mark.parametrize(
    "func, func_log", [
        # Old Submanifold
        (
            submanifold_rosenbrock(dim=20, sub_dim=8, log_transform=False),
            submanifold_rosenbrock(dim=20, sub_dim=8)
        ), 
        (
            submanifold_rosenbrock(dim=6, sub_dim=1, log_transform=False),
            submanifold_rosenbrock(dim=6, sub_dim=1)
        ),
        (
            submanifold_rosenbrock(dim=12, sub_dim=12, log_transform=False),
            submanifold_rosenbrock(dim=12, sub_dim=12) 
        ),
        # New Submanifold
        (
            SubmanifoldRosenbrock(dim=20, sub_dim=8),
            SubmanifoldRosenbrock(dim=20, sub_dim=8, log_transform=True)
        ), 
        (
            SubmanifoldRosenbrock(dim=6, sub_dim=1),
            SubmanifoldRosenbrock(dim=6, sub_dim=1, log_transform=True)
        ),
        (
            SubmanifoldRosenbrock(dim=12, sub_dim=12),
            SubmanifoldRosenbrock(dim=12, sub_dim=12, log_transform=True) 
        )
    ]
)
def test_submanifold(func, func_log):
    _test_func(func, func_log)


@pytest.mark.parametrize(
    "func, func_log", [
        (
            StyblinskyTang(dim=4),
            StyblinskyTang(dim=4, log_transform=True)
        ),
        (
            StyblinskyTang(dim=1),
            StyblinskyTang(dim=1, log_transform=True)
        ),
        (
            StyblinskyTang(dim=20),
            StyblinskyTang(dim=20, log_transform=True)
        )
    ]
)
def test_tang(func, func_log):
    _test_func(func, func_log)


@pytest.mark.parametrize(
    "func, func_log", [
        (
            Levi(),
            Levi(log_transform=True)
        )
    ]
)
def test_levi(func, func_log):
    _test_func(func, func_log)


@pytest.mark.parametrize(
    "func, func_log", [
        (
            Himmelblau(),
            Himmelblau(log_transform=True)
        )
    ]
)
def test_himmelblau(func, func_log):
    _test_func(func, func_log)


@pytest.mark.parametrize(
    "func, func_log", [
        (
            Holder(),
            Holder(log_transform=True)
        )
    ]
)
def test_holder(func, func_log):
    _test_func(func, func_log)



import time
import pytest

import random
import numpy as np
import torch

from waggon.utils import set_all_seed
from waggon.utils import fixed_random_seed
from waggon.utils import fixed_numpy_seed
from waggon.utils import fixed_torch_seed
from waggon.utils import fixed_all_seed


@pytest.mark.parametrize(
    "seed", range(20)
)
def test_set_all_seed(seed):
    set_all_seed(seed)
    # Generate for seed for first time
    x1_random = random.random()
    x1_numpy = np.random.randn(17, 23)
    x1_torch = torch.randn(13, 19)
    
    time_seed = round(time.time())
    set_all_seed(time_seed)
    set_all_seed(seed)
    
    # Generate for seed for second time
    x2_random = random.random()
    x2_numpy = np.random.randn(17, 23)
    x2_torch = torch.randn(13, 19)

    assert x1_random == x1_random
    assert np.all(x1_numpy == x2_numpy)
    assert torch.all(x1_torch == x2_torch)


@pytest.mark.parametrize(
    "seed", range(20)
)
def test_fixed_random_seed(seed):
    with fixed_random_seed(seed):
        x1 = random.random()
        y1 = random.randrange(100)
        z1 = random.uniform(0.0, 2.0)

    time_seed = round(time.time())
    set_all_seed(time_seed)

    with fixed_random_seed(seed):
        x2 = random.random()
        y2 = random.randrange(100)
        z2 = random.uniform(0.0, 2.0)

    assert x1 == x2
    assert y1 == y2
    assert z1 == z2 


@pytest.mark.parametrize(
    "seed", range(20)
)
def test_fixed_numpy_seed(seed):
    with fixed_numpy_seed(seed):
        x1 = np.random.normal(size=(25, 8))
        y1 = np.random.uniform(low=0.0, high=2.0, size=(25, 7))
        z1 = np.random.beta(1.5, 1.5, size=11)
    
    time_seed = round(time.time())
    set_all_seed(time_seed)

    with fixed_numpy_seed(seed):
        x2 = np.random.normal(size=(25, 8))
        y2 = np.random.uniform(low=0.0, high=2.0, size=(25, 7))
        z2 = np.random.beta(1.5, 1.5, size=11)

    assert np.all(x1 == x2)
    assert np.all(y1 == y2)
    assert np.all(z1 == z2)


@pytest.mark.parametrize(
    "seed", range(20)
)
def test_fixed_torch_seed(seed):
    with fixed_torch_seed(seed):
        x1 = torch.randn(18, 17)
        y1 = torch.rand(2, 3, 5)
        
    time_seed = round(time.time())
    set_all_seed(time_seed)

    with fixed_torch_seed(seed):
        x2 = torch.randn(18, 17)
        y2 = torch.rand(2, 3, 5)

    assert torch.all(x1 == x2)
    assert torch.all(y1 == y2)


@pytest.mark.parametrize(
    "seed", range(30)
)
def test_fixed_all_seed(seed):
    with fixed_all_seed(seed):
        x_random = random.random()
        x_numpy = np.random.randn()
        x_torch = torch.rand(1).item()

    time_seed = round(time.time())
    set_all_seed(time_seed)

    with fixed_all_seed(seed):
        y_random = random.random()
        y_numpy = np.random.randn()
        y_torch = torch.rand(1).item()

    assert x_random == y_random
    assert x_numpy == y_numpy
    assert x_torch == y_torch

